 /* hpgl.l -- lex file for parsing HPGL */
 /* [28-Feb-1991] */
%{

#undef yywrap		/* in case we use flex instead of lex */

#define VERSION "HPGL to PostScript translator version 1.0 [01-Mar-1991]"

#define RELATIVE_LINES

#include <stdio.h>
#include <math.h>
#include <time.h>
#include <sys/types.h>
#include <pwd.h>			/* for passwd  */

#if __STDC__
#define ARGS(plist)	plist
#else
#define ARGS(plist)	()
#endif

#define COLOR(n)	((double)(n)/(double)MAXCOLORS)
#ifdef FALSE
#undef FALSE
#endif
#define FALSE		(0)

#define MAXCOLORS	16
#define MAXLINE		255
#define PRINTF		(void)printf

#define PLOTTER_SIZE	(16.38 * 72.0)	/* longest dimension, in bp */
#define PICTURE_SIZE	(7 * 72.0)	/* standard picture size, in bp */
#define STEPS_PER_MM	40.0	/* HP7475A 8-pen plotter has 40 steps/mm */
#define SCALE(c)	((PICTURE_SIZE/PLOTTER_SIZE) * (72/25.4) * \
			(((double)c)/STEPS_PER_MM))

#ifdef TRUE
#undef TRUE
#endif
#define TRUE		(1)

#define WNA(found,expected)	do\
    {\
	PRINTF("%%? Wrong number of arguments: found %d expected %d\n",\
	    found, expected);\
	n = 0;\
	break;\
    } while (0)

#define CMD_UNKNOWN	-1

#define CMD_ABSOLUTE	0
#define CMD_COLOR	1
#define CMD_DOWN	2
#define CMD_RELATIVE	3
#define CMD_UP		4

#define MAXVALUES 4

void	do_cmd ARGS((void));
void	do_value ARGS((void));
void	init ARGS((void));

int	absolute_coordinates = TRUE;
int	cmd = CMD_UNKNOWN;
int	initialized = FALSE;
int	n = 0;
int	path_length = 0;
int	pen_is_up = TRUE;
int	values[MAXVALUES];
int	x = 0;
int	y = 0;
int	xmax = -32767;
int	ymax = -32767;
int	xmin = 32767;
int	ymin = 32767;
    
#ifdef MAX
#undef MAX
#endif
#define MAX(a,b)	(((a) > (b)) ? (a) : (b))

#ifdef MIN
#undef MIN
#endif
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))

#if defined(_IBMR2)
struct passwd	*getpwnam ARGS((char* name)); /* missing in AIX pwd.h */
#endif

int	gethostname ARGS((char *name, int namelen));
char	*getlogin ARGS((void));

%}
%%

\033\.			n = 0;

DF			n = 0;

I/[0-9 \n]		n = 0;

N			n = 0;

PD/[0-9 \n]		cmd = CMD_DOWN, n = 0;

PU/[0-9 \n]		cmd = CMD_UP, n = 0;

PA/[0-9 \n]		cmd = CMD_ABSOLUTE, n = 0;

PR/[0-9 \n]		cmd = CMD_RELATIVE, n = 0;

SC			n = 0;

SP/[0-9 \n]		cmd = CMD_COLOR, n = 0;

[0-9]+			do_value();

[;]			n = 0;	/* ignore semicolon and discard saved values */

[:, \t\n]		;	/* ignore separators */

.			{	/* comment unparsed data */
			    PRINTF("%% Unparsed: %s\n",yytext);
			    /* n = 0; */
			}
%%
void
do_cmd()
{
    int		dx;
    int		dy;

    if (!initialized)
	init();
    switch (cmd)
    {
	case CMD_ABSOLUTE:
	    absolute_coordinates = TRUE;
	    break;

	case CMD_RELATIVE:
	    absolute_coordinates = FALSE;
	    break;

	case CMD_UP:
	    pen_is_up = TRUE;
	    break;

	case CMD_DOWN:
	    pen_is_up = FALSE;
	    break;

	case CMD_COLOR:
	    if (n != 1)
		WNA(n,1);
	    else
		PRINTF("%5.3f setgray\n",COLOR(values[0]));
	    n = 0;
	    return;

	default:
	    break;
    }
    if (n < 2)
        return;
    else if (n > 2)
        WNA(n,2);
    else /* n == 2 */
    {
	if (pen_is_up)
	{
	    if (path_length > 0)
		PRINTF("S\n");
	    path_length = 0;
	    PRINTF("N\n");
	}
	else
	    path_length++;

#ifdef RELATIVE_LINES
	dx = absolute_coordinates ? (values[0] - x) : values[0];
	dy = absolute_coordinates ? (values[1] - y) : values[1];
	if (pen_is_up)
            PRINTF("%d %d M\n", x + dx, y + dy);
	else
	{		/* omit zero-length steps after first point */
	    if (!((path_length >= 2) && (dx == 0) && (dy == 0)))
                PRINTF("%d %d l\n", dx, dy);
	}


#else
        PRINTF("%d %d %s\n",values[0],values[1],
	   (pen_is_up ?
	       (absolute_coordinates ? "M" : "m") :
	       (absolute_coordinates ? "L" : "l")));
#endif

	n = 0;
	x = absolute_coordinates ? values[0] : (x + values[0]);
	y = absolute_coordinates ? values[1] : (y + values[1]);
	xmax = MAX(xmax,x);
	xmin = MIN(xmin,x);
	ymax = MAX(ymax,y);
	ymin = MIN(ymin,y);
    }
}

void
do_value()
{
	if (n >= MAXVALUES)
	{
	    PRINTF("%%?? ?too many numeric values\n");
	    n = 0;
	}
	values[n++] = atoi(yytext);
	do_cmd();
}

void
init()
{
    char	buf[MAXLINE+1];
    char	hostname[MAXLINE+1];
    struct passwd	*passwdp;
    char	*personal_name;
    time_t	timeval;	/* clock value from time() for ctime() */
    char	*user_name;

    timeval = time((time_t*)NULL);
    user_name = getlogin();
    user_name = (user_name == (char*)NULL) ? "" : user_name;
    passwdp = getpwnam(user_name);
    if (passwdp != (struct passwd *)NULL)
    {
	if (passwdp->pw_gecos[0] == '&')
	    personal_name = user_name;		/* & means use login name */
	else
	    personal_name = passwdp->pw_gecos;	/* have personal name */
    }
    if (gethostname(hostname,sizeof(hostname)) == 0)
	hostname[sizeof(hostname)-1] = '\0'; /* ensure final NULL */
    else				/* no hostname available */
	hostname[0] = '\0';

    PRINTF("%%!\n");
    PRINTF("%%%% %s\n",VERSION);

#ifdef __TIME__
#ifdef __DATE__
    PRINTF("%%%% Compiled on %s %s\n",__DATE__,__TIME__);
#endif
#endif

    PRINTF("%%%% Date: %s",ctime(&timeval));
					/* "Mon Dec 24 15:44:14 MST 1990" */
    PRINTF("%%%% User: %s <%s%s%s>\n",
	personal_name,user_name,hostname[0] ? "@" : "",hostname);
    PRINTF("%%%% Directory: %s\n",getcwd(buf,sizeof(buf)));
    PRINTF("%%%% BoundingBox: (atend)\n");
    PRINTF("/C {closepath} bind def\n");
    PRINTF("/L {lineto} bind def\n");
    PRINTF("/l {rlineto} bind def\n");
    PRINTF("/M {moveto} bind def\n");
    PRINTF("/m {rmoveto} bind def\n");
    PRINTF("/N {newpath} bind def\n");
    PRINTF("/S {stroke} bind def\n");
    PRINTF("/PICTURE_SIZE {%g} def\n",PICTURE_SIZE);
    PRINTF("/PLOTTER_SIZE {%g} def\n",PLOTTER_SIZE);
    PRINTF("PICTURE_SIZE PLOTTER_SIZE div dup scale\n");
    initialized = TRUE;
}

int
yywrap()
{
    PRINTF("%%%% BoundingBox: %d %d %d %d\n",
	(int)floor(SCALE(xmin)), (int)floor(SCALE(ymin)),
	(int)ceil(SCALE(xmax)), (int)ceil(SCALE(ymax)));
    return (1);
}
